#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include "uart.h"
#include <math.h>
#include "lcd_lib.h"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//I like these definitions
#define begin {
#define end   } 

#define int2fix(a)   (((int)(a))<<8) 
#define float2fix(a) ((int)((a)*256.0)) 
#define fix2float(a) ((double)(a)/256.0)  

#define read_bit(var,bit_num)  (var & (1<<bit_num))
#define set_bit(var,bit_num)   (var |= (1<<bit_num))
#define clear_bit(var,bit_num) (var &= ~(1<<bit_num))

#define read_flag(bit_num)  (flags & (1<<bit_num))
#define set_flag(bit_num)   (flags |= (1<<bit_num))
#define clear_flag(bit_num) (flags &= ~(1<<bit_num))

#define LED_on     (PORTD &= ~0x04)
#define LED_off    (PORTD |= 0x04)
#define LED_toggle (PORTD ^= 0x04)

#define NUM_FRAMES 16
#define FRAME_SIZE 256
#define OFFSET     150

#define N_WAVE          256    /* size of FFT */
#define LOG2_N_WAVE     8     /* log2(N_WAVE) */

#define NUM_BANKS  32
#define NUM_COEFF  15

#define squarefix(a) multfix(a,a)
#define NUM_TRAINING_FRAMES 64
#define NUM_TRAINING_FRAMES_POWER 6
#define NUM_CENTROIDS 8
#define CENTROID_POWER 3
#define NUM_SPEAKERS 3

// gets bits i, i-1 of a (0->16), used in sqrt_fix
#define get2bits(a,i) ((a >> 16-i) & 0x03)

#define pi 3.14159

// constants used for DCT
int  m1, m2, m3, m4 ;   /* 8 sample twiddle factors -- needed for 32 points */
int c1,c3,c5,c7,c9,c11,c13,c15; //16 sample factors -- needed for 32 points   
int c17,c19,c21,c23,c25,c27,c29,c31; //32 sample factors

#define float2fix(a) ((int)((a)*256.0))         //Convert float to fix. a is a float
#define fix2float(a) (((float)(a))/256.0)       //Convert fix to float. a is an int   

// Fast fixed point multiply assembler macro
#define multfix(a,b)          	  \
({                                \
int prod, val1=a, val2=b ;        \
__asm__ __volatile__ (            \ 
"muls %B1, %B2	\n\t"              \
"mov %B0, r0 \n\t"	               \ 
"mul %A1, %A2\n\t"	               \ 
"mov %A0, r1   \n\t"              \ 
"mulsu %B1, %A2	\n\t"          \ 
"add %A0, r0  \n\t"               \ 
"adc %B0, r1 \n\t"                \ 
"mulsu %B2, %A1	\n\t"          \ 
"add %A0, r0 \n\t"           \ 
"adc %B0, r1  \n\t"          \ 
"clr r1  \n\t" 		         \ 
: "=&d" (prod)               \
: "a" (val1), "a" (val2)      \
);                            \
prod;                        \
})

#define squarefix(a) multfix(a,a)

#define BUTTON_PUSHED 0
#define SOUND_DETECTED 1
#define RECENT_SOUND 2
#define SAMPLES_PROCESSED 3
#define TRAINING_MODE 4
#define PRINT_WELCOME 5

// flags used to determine state of process -------
// bit 0: button state (1 = pushed, 0 = released)
// bit 1: sound detected? (1 = yes, 0 = no)
// bit 2: recent sound detected? (1 = yes, 0 = no)
// bit 3: samples processed? (1 = yes, 0 = no)
// bit 4: training mode? (1 = yes, 0 = no)
// bit 5: print initial message? (1 = yes, 0 = no)
char flags = 0x00;

int Sinewave[N_WAVE]; // a table of sines for the FFT

// frame matrix has vertical frames
// # of columns = # of frames the sample is split into
int frame[FRAME_SIZE][NUM_FRAMES];

// coefficient matrix can store a number of sets of coefficients
int coeff[NUM_COEFF][NUM_TRAINING_FRAMES];

// number of coefficients that have been stored so far
unsigned int stored_coeffs = 0;

int c[NUM_COEFF][NUM_CENTROIDS];
int jake_centroids[NUM_COEFF][NUM_CENTROIDS];
int rishi_centroids[NUM_COEFF][NUM_CENTROIDS];
int rob_centroids[NUM_COEFF][NUM_CENTROIDS];

int speaker_dist[NUM_SPEAKERS]; // distances to each speaker
int closest; // index of the closest speaker to sample

unsigned int samples_entered = 0;

unsigned int sample_num = 3;

int speakers[NUM_SPEAKERS][NUM_COEFF][NUM_CENTROIDS] = {
// sample 0
{
{-227, 37, -204, -78, -162, 3, -104, -47},
{-210, 47, -283, -226, -247, -11, -277, -175},
{290, 43, 277, 73, 223, 6, 184, 56},
{-3415, 48, -2271, -713, -2865, -58, -1708, -404},
{16, 41, 4, -42, -70, -1, -53, -22},
{489, 33, 413, 55, 469, -2, 279, 29},
{203, 22, 225, 92, 266, 7, 220, 82},
{1137, 4, 459, 196, 761, 30, 422, 137},
{193, 20, 152, 16, 231, 4, 120, 25},
{348, 23, 409, 176, 464, 6, 444, 129},
{-133, 29, -75, 1, -117, 3, 8, 10},
{-1124, 34, -755, -358, -1119, -6, -884, -275},
{53, 30, 67, 28, 96, 5, 132, 30},
{-89, 34, 81, 128, -95, 5, 151, 83},
{-59, 32, -36, -7, -118, 4, -49, -7}
},
// sample 1
{
{108, -96, 4, -61, 35, -65, 6, -39},
{113, -584, -7, -298, -27, -640, -63, -119},
{160, 284, 2, 66, 37, 211, 6, 42},
{-6, -3190, -23, -658, -76, -2334, -105, -394},
{142, -51, 0, -60, 22, -124, -10, -13},
{187, 275, -6, 21, 6, 190, -40, 26},
{168, 438, 4, 134, 57, 433, 16, 53},
{57, 1948, 26, 285, 130, 1341, 102, 151},
{158, 159, 4, 19, 46, 133, -2, 12},
{179, 775, 1, 230, 53, 785, 5, 90},
{124, 139, 5, 33, 48, 156, 12, 8},
{174, -2409, 3, -521, -16, -2064, -53, -201},
{130, 220, 5, 45, 51, 223, 8, 17},
{105, 551, 0, 202, 40, 587, 21, 73},
{94, 32, 2, -1, 27, 6, 6, 4}
},
// sample 2
{
{28, -171, 8, -68, 23, -108, 1, -7},
{37, -501, 5, -186, 14, -523, -29, -286},
{39, 221, 7, 56, 27, 146, 16, 26},
{53, -2360, -3, -830, -70, -1406, -85, -295},
{42, -87, 5, -57, 16, -92, 2, -40},
{44, 216, 5, 100, 29, 59, -1, -156},
{39, 237, 6, 109, 29, 238, 16, 63},
{40, 898, 19, 232, 32, 633, 58, 342},
{37, 46, 5, 53, 29, 40, 3, -26},
{39, 518, 8, 211, 40, 402, 16, 6},
{35, 16, 7, -4, 26, 47, 4, 27},
{37, -1450, 5, -438, 28, -1036, -53, -208},
{32, 125, 7, 43, 30, 78, 6, -6},
{31, 302, 8, 82, 23, 312, 15, 85},
{26, -70, 5, -29, 16, -20, 2, 16}
}
};

#define BUTTON_CHECK_MS 50
#define SOUND_DETECT_MS 20
#define MESSAGE_PRINT_MS 700

volatile char button_check_timer = BUTTON_CHECK_MS;
volatile char sound_detected_timer = SOUND_DETECT_MS;
volatile unsigned int message_print_timer = MESSAGE_PRINT_MS;

char hamming_256[256] = {
0b00010100, 0b00010100, 0b00010100, 0b00010100, 0b00010101, 0b00010101, 0b00010101, 0b00010110,
0b00010110, 0b00010111, 0b00011000, 0b00011000, 0b00011001, 0b00011010, 0b00011011, 0b00011100,
0b00011101, 0b00011110, 0b00011111, 0b00100001, 0b00100010, 0b00100011, 0b00100101, 0b00100110,
0b00101000, 0b00101010, 0b00101011, 0b00101101, 0b00101111, 0b00110001, 0b00110011, 0b00110101,
0b00110111, 0b00111001, 0b00111011, 0b00111101, 0b00111111, 0b01000010, 0b01000100, 0b01000110,
0b01001001, 0b01001011, 0b01001110, 0b01010000, 0b01010011, 0b01010101, 0b01011000, 0b01011011,
0b01011101, 0b01100000, 0b01100011, 0b01100101, 0b01101000, 0b01101011, 0b01101110, 0b01110001,
0b01110011, 0b01110110, 0b01111001, 0b01111100, 0b01111111, 0b10000010, 0b10000101, 0b10001000,
0b10001010, 0b10001101, 0b10010000, 0b10010011, 0b10010110, 0b10011001, 0b10011100, 0b10011111,
0b10100010, 0b10100100, 0b10100111, 0b10101010, 0b10101101, 0b10110000, 0b10110010, 0b10110101,
0b10111000, 0b10111010, 0b10111101, 0b11000000, 0b11000010, 0b11000101, 0b11000111, 0b11001010,
0b11001100, 0b11001110, 0b11010001, 0b11010011, 0b11010101, 0b11010111, 0b11011010, 0b11011100,
0b11011110, 0b11100000, 0b11100010, 0b11100100, 0b11100101, 0b11100111, 0b11101001, 0b11101011,
0b11101100, 0b11101110, 0b11101111, 0b11110001, 0b11110010, 0b11110011, 0b11110101, 0b11110110,
0b11110111, 0b11111000, 0b11111001, 0b11111010, 0b11111011, 0b11111100, 0b11111100, 0b11111101,
0b11111101, 0b11111110, 0b11111110, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111,
0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111111, 0b11111110, 0b11111110, 0b11111101,
0b11111101, 0b11111100, 0b11111100, 0b11111011, 0b11111010, 0b11111001, 0b11111000, 0b11110111,
0b11110110, 0b11110101, 0b11110011, 0b11110010, 0b11110001, 0b11101111, 0b11101110, 0b11101100,
0b11101011, 0b11101001, 0b11100111, 0b11100101, 0b11100100, 0b11100010, 0b11100000, 0b11011110,
0b11011100, 0b11011010, 0b11010111, 0b11010101, 0b11010011, 0b11010001, 0b11001110, 0b11001100,
0b11001010, 0b11000111, 0b11000101, 0b11000010, 0b11000000, 0b10111101, 0b10111010, 0b10111000,
0b10110101, 0b10110010, 0b10110000, 0b10101101, 0b10101010, 0b10100111, 0b10100100, 0b10100010,
0b10011111, 0b10011100, 0b10011001, 0b10010110, 0b10010011, 0b10010000, 0b10001101, 0b10001010,
0b10001000, 0b10000101, 0b10000010, 0b01111111, 0b01111100, 0b01111001, 0b01110110, 0b01110011,
0b01110001, 0b01101110, 0b01101011, 0b01101000, 0b01100101, 0b01100011, 0b01100000, 0b01011101,
0b01011011, 0b01011000, 0b01010101, 0b01010011, 0b01010000, 0b01001110, 0b01001011, 0b01001001,
0b01000110, 0b01000100, 0b01000010, 0b00111111, 0b00111101, 0b00111011, 0b00111001, 0b00110111,
0b00110101, 0b00110011, 0b00110001, 0b00101111, 0b00101101, 0b00101011, 0b00101010, 0b00101000,
0b00100110, 0b00100101, 0b00100011, 0b00100010, 0b00100001, 0b00011111, 0b00011110, 0b00011101,
0b00011100, 0b00011011, 0b00011010, 0b00011001, 0b00011000, 0b00011000, 0b00010111, 0b00010110,
0b00010110, 0b00010101, 0b00010101, 0b00010101, 0b00010100, 0b00010100, 0b00010100, 0b00010100
};

char bank_start_indices_32[32] = {6, 7, 9, 11, 12, 14, 16, 18, 20, 22, 25, 27, 30, 33, 35, 38, 42, 45, 48, 52, 56, 60, 64, 69, 73, 78, 84, 89, 95, 101, 107, 114};

char max_bank_size = 13;

char banks_32[32][13] = {
{0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b10000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b10000000,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b10101010,0b01010101,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01010101,0b10101010,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b11001100,0b10011001,0b01100110,0b00110011,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b00110011,0b01100110,0b10011001,0b11001100,0b11111111,0b11000000,0b10000000,0b01000000,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b01000000,0b10000000,0b11000000,0b11111111,0b11001100,0b10011001,0b01100110,0b00110011,0b00000000,0b00000000,0b00000000,0b00000000,0b00000000},
{0b00110011,0b01100110,0b10011001,0b11001100,0b11111111,0b11010101,0b10101010,0b10000000,0b01010101,0b00101010,0b00000000,0b00000000,0b00000000},
{0b00101010,0b01010101,0b10000000,0b10101010,0b11010101,0b11111111,0b11001100,0b10011001,0b01100110,0b00110011,0b00000000,0b00000000,0b00000000},
{0b00110011,0b01100110,0b10011001,0b11001100,0b11111111,0b11010101,0b10101010,0b10000000,0b01010101,0b00101010,0b00000000,0b00000000,0b00000000},
{0b00101010,0b01010101,0b10000000,0b10101010,0b11010101,0b11111111,0b11010101,0b10101010,0b10000000,0b01010101,0b00101010,0b00000000,0b00000000},
{0b00101010,0b01010101,0b10000000,0b10101010,0b11010101,0b11111111,0b11010101,0b10101010,0b10000000,0b01010101,0b00101010,0b00000000,0b00000000},
{0b00101010,0b01010101,0b10000000,0b10101010,0b11010101,0b11111111,0b11011011,0b10110110,0b10010010,0b01101101,0b01001001,0b00100100,0b00000000},
{0b00100100,0b01001001,0b01101101,0b10010010,0b10110110,0b11011011,0b11111111,0b11011011,0b10110110,0b10010010,0b01101101,0b01001001,0b00100100},
{0b00100100,0b01001001,0b01101101,0b10010010,0b10110110,0b11011011,0b11111111,0b11011011,0b10110110,0b10010010,0b01101101,0b01001001,0b00100100}
};

int8_t lcd_buffer[16];	// LCD display buffer

//=======================================================
//timer 0 overflow ISR, runs every 1ms
ISR (TIMER0_COMPA_vect) 
begin
  //Decrement the time if not already zero
  if (button_check_timer>0) --button_check_timer;
  if (message_print_timer>0) --message_print_timer;
  
  if (read_flag(SOUND_DETECTED)) {
  	sound_detected_timer = SOUND_DETECT_MS;
	set_flag(RECENT_SOUND);
	clear_flag(SOUND_DETECTED);
  }
  else if (sound_detected_timer > 0) --sound_detected_timer;
  else clear_flag(RECENT_SOUND);

end

//comparator sound detected interrupt
ISR(ANALOG_COMP_vect)
{
	set_flag(SOUND_DETECTED);
}

//=======================================================

// returns the index of the smallest element of an array
// called by methods getCetroids() and recognize()
unsigned int min_abs_index(int a[], unsigned int num_elements) {
	int smallest = 32767; // start at max int value
	unsigned int index = 0; // index of the smallest element
	for (int i=0; i<num_elements; i++) {
		if (a[i] < 0) {
			a[i] = multfix(a[i],int2fix(-1));
		}
		if (a[i] < smallest) {
			smallest = a[i];
			index = i;
		}
	}
	return index;
}

// returns the index of the smallest element of an array
// called by method recognize()
int min(int a[], unsigned int num_elements) {
	int smallest = 32767; // start at max int value
	unsigned int index = 0; // index of the smallest element
	for (int i=0; i<num_elements; i++) {
		if (a[i] < smallest) {
			smallest = a[i];
			index = i;
		}
	}
	return smallest;
}

// Turns FFT frame into filter bank coefficients frame
void filterBank(void) {
	for (unsigned int i=0; i<NUM_FRAMES; i++) {
		for (unsigned int j=0; j<NUM_BANKS; j++) {
			int bank_total = 0;
			int k = bank_start_indices_32[j] + 1;
			int m = 0;
			while (banks_32[j][m] != 0) {
				bank_total += multfix(frame[k][i],(int)banks_32[j][m]);
				m++;
				k++;
			}
			frame[j][i] = bank_total; // watch for overflow
		}
		for (unsigned int j=NUM_BANKS; j<FRAME_SIZE; j++) {
			frame[j][i] = 0; // clear frame entries after banks
		}
	}
}

// uses FFTfix() to perform fixed point FFT of the columns of a matrix
// changes matrix to store magnitudes of complex results
void FFTframes(int m) {
	int fr[FRAME_SIZE];
	int fi[FRAME_SIZE];
	for (int i=0; i<NUM_FRAMES; i++) {
		// do FFT on a column
		for (int k=0; k<FRAME_SIZE; k++) {
			fr[k] = frame[k][i];
			fi[k] = 0;
		}
		FFTfix(fr,fi,m);
		// set column to magnitude values
		for (int j=0; j<FRAME_SIZE; j++) {
			frame[j][i] = squarefix(fr[j]) + squarefix(fi[j]); // give square FFT vals?!!!!!!!!!!!!!!!!!!!!!!
			//frame[j][i] = fr[j];
		}
	}
}

// Perform 32pt DCT on filter bank coefficients stored in frames
void DCTframes(void) {
	int a32[32];  // the 32 sample inputs
	int S32[32];  // the 32 sample outputs
	for (unsigned int i = 0; i<NUM_FRAMES; i++) {
		// do DCT to a column (only to # banks)
		for (int k=0; k<NUM_BANKS; k++) {
			a32[k] = frame[k][i];
			frame[k][i] = 0;
		}
		DCT1D_32(a32,S32);
		for (int k=0; k<NUM_COEFF; k++) {
			frame[k][i] = S32[k+1];
		}
	}
}

// write centroids to EEprom to save RAM
// number of centroids is 2^centroid_power
void getCentroids() {
	unsigned int d_tot = 0; // avg distance to centroids
	unsigned int d_tot_last = 0; // last avg distance to centroids
	// DEBUG -- keep d_tot history -----------------------------------------
	//unsigned int d_tot_history[CENTROID_POWER+1]; // distance for each num centroids
	//unsigned int d_tot_history_counter = 0;
	// ---------------------------------------------------------------------
	unsigned int n_x[NUM_TRAINING_FRAMES]; // index of corresponding centroid for each sample vector
	int d_x[NUM_CENTROIDS]; // distances from a single sample to all centroids
	int d= 0; // a placeholder for distance
	int e = 0b0000000000001000; // set error
	int c_num = 1; // current number of centroids
	char new_c = 0; // flag indicating whether new centroids have been added
	int num_loops = 0; // indicates number of cycles through inner error loop

	// find starting center point
	for (int i=0; i<NUM_TRAINING_FRAMES; i++) {
		for (int j=0; j<NUM_COEFF; j++) {
			c[j][0] += coeff[j][i]; // divide by num training frames
		}
	}
	// divide by 16
	for (int i=0; i<NUM_COEFF; i++) {
		c[i][0] >>= NUM_TRAINING_FRAMES_POWER; // divide by num training frames
	}
	// find avg distance to the new center point
	for (int i=0; i<NUM_TRAINING_FRAMES; i++) {
		for (int j=0; j<NUM_COEFF; j++) {
			d = coeff[j][i]-c[j][0];
			if (d<0) {
				d = multfix(d,int2fix(-1));
			}
			d_tot += d>>4; // divide to prevent overflow
			if (d_tot < 0) {
				fprintf(stdout,"overflow occurred\r\n");
			}
		}
	}
	// DEBUG -- record d_tot in history -------------
	//d_tot_history[d_tot_history_counter] = d_tot;
	//d_tot_history_counter++;
	// ----------------------------------------------
	
	while (c_num < NUM_CENTROIDS) {
		// split each centroid
		for (int i=0; i<c_num; i++) {
			for (int j=0; j<NUM_COEFF; j++) {
				c[j][i] = multfix(int2fix(1)+e,c[j][i]);
				c[j][i+c_num] = multfix(int2fix(1)-e,c[j][i]);
			}
		}
		c_num <<= 1; // multiply c_num by 2;
		new_c = 1;
		num_loops = 0;
		while ((new_c || d_tot_last-d_tot > e) && num_loops<15) {
			new_c = 0; // clear flag
			num_loops++;
		 
			// find distances
			for (int i=0; i<NUM_TRAINING_FRAMES; i++) {
				for (int j=0; j<c_num; j++) {
					d_x[j] = 0; // initialize distance to 0
					for (int k=0; k<NUM_COEFF; k++) {
						d_x[j] += coeff[k][i]-c[k][j];
					}
				}
				n_x[i] = min_abs_index(d_x, c_num); // set index of min distance for each x
			}
			unsigned int count;
			// update codevectors
			for (int i=0; i<c_num; i++) {
				// clear c
				for (int k=0; k<NUM_COEFF; k++) {
					c[k][i] = 0;
				}
				count = 0; // count of samples for each c
				for (int j=0; j<NUM_TRAINING_FRAMES; j++) {
					if (n_x[j] == i) {
						for (int k=0; k<NUM_COEFF; k++) {
							c[k][i] += coeff[k][j];
						}
						count++;
					}
				}
				if (count > 0) {
					for (int k=0; k<NUM_COEFF; k++) {
						c[k][i] = divfix(c[k][i],int2fix(count));
					}
				}
			}
			d_tot_last = d_tot;
			d_tot = 0;
			int d;
			for (int i=0; i<NUM_TRAINING_FRAMES; i++) {
				for (int k=0; k<NUM_COEFF; k++) {
					d = coeff[k][i]-c[k][n_x[i]];
					if (d<0) d = multfix(d,int2fix(-1));
					d_tot += d>>4; // divide to prevent overflow
				}
			}
			// DEBUG -- record d_tot in history -------------
			//d_tot_history[d_tot_history_counter] = d_tot;
			//d_tot_history_counter++;
			// ----------------------------------------------
		}
	}
}


// get the index of the closest speaker
void recognize(void) {
	int min_dist; // sum of distances from each sample to its closest centroid
	unsigned int coeff_dist[NUM_CENTROIDS]; // distances of one sample to each centroid
	int difference;	
	for (int i=0; i<NUM_SPEAKERS; i++) {
		min_dist = 0;
		for (int j=0; j<NUM_CENTROIDS; j++) { // sample centroids
		
			// calculates euclidean distance
			for (int k=0; k<NUM_CENTROIDS; k++) { // saved centroids
				coeff_dist[k] = 0;
				for (int l=0; l<NUM_COEFF; l++) {
					//difference = c[l][j] - speakers[i][l][k];
					//if (difference<0) difference = multfix(difference,int2fix(-1));
					//coeff_dist[k] += difference;
					coeff_dist[k] += squarefix(c[l][j]>>2 - speakers[i][l][k]>>2);
				}
				coeff_dist[k] = sqrt_fix(coeff_dist[k]);
			}
			min_dist += min(coeff_dist, NUM_CENTROIDS);
		}
		speaker_dist[i] = min_dist;
	}
	//fprintf(stdout,"speaker_dist[%i]: %i\r\n",i,speaker_dist[i]);
}

// fixed point sqrt function using the shifting nth root algorithm
// optimized for binary square root
int sqrt_fix(int input) {
	int alpha = 0;
	int y = 0;
	int r = 0;
	int i = 0;
	int i_initial = 0;
	int beta_eq = 0;
	
	if (input == 0) return 0;

	// skip past leading zeros
	while (alpha == 0) {
		i += 2;
		alpha = get2bits(input,i);
	}
	if (alpha == 0) {
		return 0;
	}
	i_initial = i;
	
	// start algorithm
	for (i=i_initial; i<=24; i+=2) {
		if (i<=16) alpha = get2bits(input,i);
		else alpha = 0;
		beta_eq = 1+((y-r) << 2);
		if (beta_eq < alpha) { // beta is 1
			r = alpha - beta_eq;
			y <<= 1;
			y++;
		}
		else { // beta is 0
			r <<= 2;
			r += alpha;
			y <<= 1;
		}
	}
	return y;
}

// Bruce Land's code to divide fixed point numbers
int divfix(int nn, int dd)
begin
    int x, d ;
    signed char count, neg ;
    count = 0;
    neg = 0 ; 
    d = dd ;
    
    // only works with + numbers
    if (d & 0x8000)
    begin
        neg = 1;
        d = -d ;
    end
    
    // range reduction
    while (d>0x0100)
    begin
        --count ;
        d >>= 1 ;
    end
    
    while (d<0x0080)
    begin
        ++count ;
        d <<= 1 ;
    end
     
    // Newton interation
    x = 0x02ea - (d<<1) ;
    x = multfix(x, 0x0200-multfix(d,x));
    x = multfix(x, 0x0200-multfix(d,x)); 

    
    // range expansion
    if (count>0)  x = x<<count ;
    else if (count<0) x = x>>(-count) ;  
    
    // fix sign
    if (neg==1) x=-x;
    
    //form ratio
    x = multfix(x,nn) ;  
     
    return x ;  
end

//=======================================================

void FFTfix(int fr[], int fi[], int m)
//Adapted from code by:
//Tom Roberts 11/8/89 and Malcolm Slaney 12/15/94 malcolm@interval.com
//fr[n],fi[n] are real,imaginary arrays, INPUT AND RESULT.
//size of data = 2**m 
// This routine does foward transform only
begin 
    int mr,nn,i,j,L,k,istep, n;
    int qr,qi,tr,ti,wr,wi; 
    
    mr = 0;
    n = 1<<m;   //number of points
    nn = n - 1;

    /* decimation in time - re-order data */
    for(m=1; m<=nn; ++m) 
    begin
        L = n;
        do L >>= 1; while(mr+L > nn);
        mr = (mr & (L-1)) + L;
        if(mr <= m) continue;
        tr = fr[m];
        fr[m] = fr[mr];
        fr[mr] = tr;
        //ti = fi[m];   //for real inputs, don't need this
        //fi[m] = fi[mr];
        //fi[mr] = ti;
    end
    L = 1;
    k = LOG2_N_WAVE-1;
    while(L < n) 
    begin
        istep = L << 1;
        for(m=0; m<L; ++m) 
        begin
            j = m << k;
            wr =  Sinewave[j+N_WAVE/4];
            wi = -Sinewave[j];
            wr >>= 1;
            wi >>= 1;
            
            for(i=m; i<n; i+=istep) 
            begin
                j = i + L;
                tr = multfix(wr,fr[j]) - multfix(wi,fi[j]);
                ti = multfix(wr,fi[j]) + multfix(wi,fr[j]);
                qr = fr[i] >> 1;
                qi = fi[i] >> 1;     
                fr[j] = qr - tr;
                fi[j] = qi - ti;
                fr[i] = qr + tr;
                fi[i] = qi + ti;
            end
        end
        --k;
        L = istep;
    end
end

//==================================================================  
void DCT1D_02(int a[], int S[])
begin
/*
Modified for GCC by Bruce Land BRL4@cornell.edu
DCT1D_02.c
version mar jan  1 08:45:57 GMT 2002
whygee@f-cpu.org

1* 8-bin DCT for a "baseline" JPG compressor.

originally cut and pasted from : sbcci_DCT2D.pdf
"Pipelined Fast 2-D DCT Architecture for JPEG Image Compression"
Luciano Volcan Agostini  <agostini@inf.ufrgs.br>
Ivan Saraiva Silva       <ivan@dimap.ufrn.br>
Sergio Bampi             <bampi@inf.ufrgs.br>
*/
    int b0, b1, b2, b3, b4, b5, b6, b7;
    int c0, c1, c2, c3, c4, c5, c6;
    int d0, d1, d3, d4;
    int e2, e3, e4, e6, e7;
    int f2, f3, f4, f5, f6, f7;
    
    /* Step 1 */
    b0 = a[0] + a[7];
    b1 = a[1] + a[6];
    b2 = a[3] - a[4]; /* corrected */
    b3 = a[1] - a[6];
    b4 = a[2] + a[5];
    b5 = a[3] + a[4];
    b6 = a[2] - a[5];
    b7 = a[0] - a[7];  /***/
    
    /* Step 2 */
    c0 = b0 + b5;
    c1 = b1 - b4;
    c2 = b2 + b6;  /***/
    c3 = b1 + b4;
    c4 = b0 - b5;
    c5 = b3 + b7;  /***/
    c6 = b3 + b6;  /***/
    
    /* Step 3 */
    d0 = c0 + c3;  /***/
    d1 = c0 - c3;  /***/
    d3 = c1 + c4;
    d4 = c2 - c5;
    
    /* Step 4 */
    e2 = multfix(m3,c2);  /*c2*/
    e3 = multfix(m1,c6);  /*c6*/
    e4 = multfix(m4,c5);  /*c5*/
    e6 = multfix(m1,d3);
    e7 = multfix(m2,d4);
    
    /* Step 5 */
    f2 = c4 + e6;  /*c4*/
    f3 = c4 - e6;  /*c4*/
    f4 = e3 + b7;  /*b7*/
    f5 = b7 - e3;  /*b7*/
    f6 = e2 + e7;
    f7 = e4 + e7;
    
    /* Step 6 */
    S[0] = d0;       /*d0*/
    S[1] = f4 + f7;
    S[2] = f2;
    S[3] = f5 - f6;
    S[4] = d1;       /*d1*/
    S[5] = f5 + f6;
    S[6] = f3;
    S[7] = f4 - f7;
end

//=========================================================  
// From the algorithm to combine 2 8-point DCTs given in
// KR Rao and P Yip:
// "Discrete Cosine Transform", Academic Press 1990
// pages 60-61
void DCT1D_16(int a16[], int S16[])
begin
    //segment 16 samples into first 8 
    //then call DCT1D_02 
    //and assign outputs to correct frequencies
	int a[8]; // 8-pt DCT input
	int S[8]; // 8-pt DCT output
    a[0] = a16[0] + a16[15];
    a[1] = a16[1] + a16[14];
    a[2] = a16[2] + a16[13];
    a[3] = a16[3] + a16[12];
    a[4] = a16[4] + a16[11];
    a[5] = a16[5] + a16[10];
    a[6] = a16[6] + a16[9];
    a[7] = a16[7] + a16[8];
    DCT1D_02(a,S); 
    S16[0] = S[0];
    S16[2] = S[1];
    S16[4] = S[2];
    S16[6] = S[3]; 
    S16[8] = S[4];
    S16[10] = S[5];
    S16[12] = S[6];
    S16[14] = S[7];
    
    //segment 16 samples into second 8 
    //premult by factors 
    //then call DCT1D_02 
    //and assign outputs to correct frequencies
    a[0] = multfix(c1, (a16[0] - a16[15]));
    a[1] = multfix(c3, (a16[1] - a16[14]));
    a[2] = multfix(c5, (a16[2] - a16[13]));
    a[3] = multfix(c7, (a16[3] - a16[12])); 
    a[4] = multfix(c9, (a16[4] - a16[11]));
    a[5] = multfix(c11, (a16[5] - a16[10]));
    a[6] = multfix(c13, (a16[6] - a16[9]));
    a[7] = multfix(c15, (a16[7] - a16[8]));
    DCT1D_02(a,S); 
    S16[1] = S[0] + S[1];
    S16[3] = S[1] + S[2];
    S16[5] = S[2] + S[3];
    S16[7] = S[3] + S[4]; 
    S16[9] = S[4] + S[5];
    S16[11] = S[5] + S[6];
    S16[13] = S[6] + S[7];
    S16[15] = S[7] ;    
end

//=========================================================  
// From the algorithm to combine 2 16-point DCTs given in
// KR Rao and P Yip:
// "Discrete Cosine Transform", Academic Press 1990
// pages 60-61
void DCT1D_32(int a32[], int S32[])
begin
    //segment 16 samples into first 16 
    //then call DCT1D_16 
    //and assign outputs to correct frequencies
	int a16[16]; // 16-pt DCT input
	int S16[16]; // 16-pt DCT output
    a16[0] = a32[0] + a32[31];
    a16[1] = a32[1] + a32[30];
    a16[2] = a32[2] + a32[29];
    a16[3] = a32[3] + a32[28];
    a16[4] = a32[4] + a32[27];
    a16[5] = a32[5] + a32[26];
    a16[6] = a32[6] + a32[25];
    a16[7] = a32[7] + a32[24];
    a16[8] = a32[8] + a32[23];
    a16[9] = a32[9] + a32[22];
    a16[10] = a32[10] + a32[21];
    a16[11] = a32[11] + a32[20];
    a16[12] = a32[12] + a32[19];
    a16[13] = a32[13] + a32[18];
    a16[14] = a32[14] + a32[17];
    a16[15] = a32[15] + a32[16];
    DCT1D_16(a16,S16); 
    S32[0] = S16[0];
    S32[2] = S16[1];
    S32[4] = S16[2];
    S32[6] = S16[3]; 
    S32[8] = S16[4];
    S32[10] = S16[5];
    S32[12] = S16[6];
    S32[14] = S16[7];
    S32[16] = S16[8];
    S32[18] = S16[9];
    S32[20] = S16[10];
    S32[22] = S16[11]; 
    S32[24] = S16[12];
    S32[26] = S16[13];
    S32[28] = S16[14];
    S32[30] = S16[15];
    
    //segment 16 samples into second 16 
    //premult by factors 
    //then call DCT1D_16
    //and assign outputs to correct frequencies
    a16[0] = multfix(c1, (a32[0] - a32[31]));
    a16[1] = multfix(c3, (a32[1] - a32[30]));
    a16[2] = multfix(c5, (a32[2] - a32[29]));
    a16[3] = multfix(c7, (a32[3] - a32[28])); 
    a16[4] = multfix(c9, (a32[4] - a32[27]));
    a16[5] = multfix(c11, (a32[5] - a32[26]));
    a16[6] = multfix(c13, (a32[6] - a32[25]));
    a16[7] = multfix(c15, (a32[7] - a32[24]));
    a16[8] = multfix(c17, (a32[8] - a32[23]));
    a16[9] = multfix(c19, (a32[9] - a32[22]));
    a16[10] = multfix(c21, (a32[10] - a32[21]));
    a16[11] = multfix(c23, (a32[11] - a32[20])); 
    a16[12] = multfix(c25, (a32[12] - a32[19]));
    a16[13] = multfix(c27, (a32[13] - a32[18]));
    a16[14] = multfix(c29, (a32[14] - a32[17]));
    a16[15] = multfix(c31, (a32[15] - a32[16]));
    DCT1D_16(a16,S16); 
    S32[1] = S16[0] + S16[1];
    S32[3] = S16[1] + S16[2];
    S32[5] = S16[2] + S16[3];
    S32[7] = S16[3] + S16[4]; 
    S32[9] = S16[4] + S16[5];
    S32[11] = S16[5] + S16[6];
    S32[13] = S16[6] + S16[7];
    S32[15] = S16[7] + S16[8];
    S32[17] = S16[8] + S16[9];
    S32[19] = S16[9] + S16[10];
    S32[21] = S16[10] + S16[11];
    S32[23] = S16[11] + S16[12]; 
    S32[25] = S16[12] + S16[13];
    S32[27] = S16[13] + S16[14];
    S32[29] = S16[14] + S16[15];
    S32[31] = S16[15] ;    
end
//=========================================================

void readButton(void) {
	// new button push? check button state and the pin
	if (!read_flag(BUTTON_PUSHED) && !read_bit(PINB, 0)) {
		set_flag(BUTTON_PUSHED);
	}
	// button release? check button state and the pin
	if (read_flag(BUTTON_PUSHED) && read_bit(PINB, 0)) {
	 	clear_flag(BUTTON_PUSHED);
	}
}

void collectSamples(void) {
	unsigned int limit = NUM_FRAMES*OFFSET+FRAME_SIZE-OFFSET;
	unsigned int j = 0;
	unsigned int k = 0;
	unsigned int jx = 0;
	unsigned int kx = 0;
	int converted;

	// initialize frame matrix to 0's
	for (j=0; j<FRAME_SIZE; j++) {
		for (k=0; k<NUM_FRAMES; k++) {
			frame[j][k] = 0;
		}
	}

	j = 0;
	k = 0;
	for (unsigned int i=0; i<limit; i++) {
		ADCSRA |= (1<<ADSC); // start conversion
		while (ADCSRA & (1<<ADSC)); // do nothing
		converted = ADCH;
		if (converted == 0) converted = 1; // 0 represents unfilled
		
		/* DEBUG write to samples array ---
		sample[i] = converted;
		// ----------------------------- */

		// set current frame location to ADCH
		frame[j][k] = converted<<6; // shift left by scaling factor
		
		// add other required copies
		jx = j;
		kx = k;
		while(jx >= OFFSET && kx < NUM_FRAMES-1) {
			jx = jx - OFFSET;
			kx = kx + 1;
			frame[jx][kx] = converted<<6; // shift left by scaling factor
		}
		
		// increment j,k
		while(frame[j][k] && i<limit-1) {
			j = (j+1)%FRAME_SIZE; // j is overflows w/ frame size
			if (!j) k++; // k increments when j overflows
		}
	}
}

void initializeAtoD(void) {
	// set Aref to Vcc, read high byte only, use input A0
	ADMUX = (1<<REFS0)|(1<<ADLAR);
	// enable AD converter
	ADCSRA = (1<<ADEN); 
	// set prescalar to 125kHz
	ADCSRA += 7;
	// set port A0 to input
	DDRA = (0<<PINA0); 
}

void initializeSquareWave(void) {
	// set up timer 2 for square wave with NO ISR
	// 200 cycle period from OC2
	OCR2A = 29;
	//  count at full rate
	TCCR2B =  1;	
	// set to toggle OC2A, clear on match,
	TCCR2A = (1<<COM2A0) | (1<<WGM21) ;
	// PORT D.7 is OC2A
	DDRD = (1<<PIND7) ;
}

void initializeButtonTimer(void) {
	//set up timer 0 for 16 mSec timebase 
	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
	OCR0A = 249;  		//set the compare re to 250 time ticks
	//set prescalar to divide by 64
	TCCR0B |= (1<<CS01) | (1<<CS00);
	// turn on clear-on-match
	TCCR0A= (1<<WGM01) ;
}

// initialize constants for DCT
void initializeDCT(void) {
  //constants for 8 sample DCT [also used by 16 sample)
  m1 = float2fix(cos(4*pi/16));
  m2 = float2fix(cos(6*pi/16));
  m3 = float2fix(cos(2*pi/16) - cos(6*pi/16));
  m4 = float2fix(cos(2*pi/16) + cos(6*pi/16));
  // constants for 16 bit DCT 
  c1 = float2fix(0.5*cos(1.0*pi/32.0));
  c3 = float2fix(0.5*cos(3.0*pi/32.0));
  c5 = float2fix(0.5*cos(5.0*pi/32.0));
  c7 = float2fix(0.5*cos(7.0*pi/32.0));
  c9 = float2fix(0.5*cos(9.0*pi/32.0));
  c11 = float2fix(0.5*cos(11.0*pi/32.0));
  c13 = float2fix(0.5*cos(13.0*pi/32.0));
  c15 = float2fix(0.5*cos(15.0*pi/32.0)); 
  c17 = float2fix(0.5*cos(17.0*pi/32.0));
  c19 = float2fix(0.5*cos(19.0*pi/32.0));
  c21= float2fix(0.5*cos(21.0*pi/32.0));
  c23= float2fix(0.5*cos(23.0*pi/32.0));
  c25= float2fix(0.5*cos(25.0*pi/32.0));
  c27 = float2fix(0.5*cos(27.0*pi/32.0));
  c29 = float2fix(0.5*cos(29.0*pi/32.0));
  c31 = float2fix(0.5*cos(31.0*pi/32.0)); 
}

void init_lcd(void) 
begin
	LCDinit();	//initialize the display
	LCDcursorOFF();
	LCDclr();	//clear the display
	LCDGotoXY(0,0);
end	

//========================================================         

void process_samples(void) {
	collectSamples();
	//---------------------- HAMMING -----------------------
	for (unsigned int i=0; i<NUM_FRAMES; i++) {
		for (unsigned int j=0; j<FRAME_SIZE; j++) {
			frame[j][i] = multfix(frame[j][i],(int)hamming_256[j]);
		}
	}
	// ----------------------- FFT -------------------------
	FFTframes(LOG2_N_WAVE);
	// -------------------- MEL BANKS ----------------------
	filterBank();
	// ----------------------- DCT -------------------------
	DCTframes();
}

// performs a mod b
unsigned int mod(unsigned int a, unsigned int b) {
	if (a>=b) return mod(a-b,b);
	else return a;
}

int main(void)
begin  

	initializeSquareWave();
	initializeButtonTimer();
	initializeAtoD();
	initializeDCT();
	init_lcd();
	sei();

	// set onboard LED port to output
	DDRD |= 0x04;

	// set port B to inputs with pull-ups on (for pushbutton, can change)
	DDRB = 0x00 ; // input
	PORTB = 0xff; // pullups on
	
	DDRA = 0x74 ; // output on pins 2,4,5,6
	PORTA = 0x70; // pullups on for LEDs

	// Enable analog comp
	ACSR = (1<<ACIE) ;

	//init the UART -- uart_init() is in uart.c
	uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"Starting...\n\r");  
  
	LED_off;

	// one cycle sine table
	//  required for FFT
	for (unsigned int i=0; i<N_WAVE; i++) Sinewave[i] = float2fix(sin(6.283*((float)i)/N_WAVE)); 
	
	set_flag(PRINT_WELCOME);

	// main loop
	while(1) {
		if (read_flag(PRINT_WELCOME)) {
			LCDclr();
			sprintf(lcd_buffer,"Hold the button");	                 
			LCDGotoXY(0,0);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			sprintf(lcd_buffer,"and speak into");	                 
			LCDGotoXY(0,1);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			sprintf(lcd_buffer,"the microphone");	                 
			LCDGotoXY(-4,2);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			sprintf(lcd_buffer,"when ready");	                 
			LCDGotoXY(-4,3);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			clear_flag(PRINT_WELCOME);
		}

		if (button_check_timer == 0) {
			readButton();
			button_check_timer = BUTTON_CHECK_MS;
		}

		if (message_print_timer == 1) {
			set_flag(PRINT_WELCOME);
		}

		if (read_bit(PINB, 5) && !read_flag(TRAINING_MODE)) {
			set_flag(TRAINING_MODE);
			LCDclr();
			sprintf(lcd_buffer,"Training Mode!");	                 
			LCDGotoXY(1,1);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			message_print_timer = MESSAGE_PRINT_MS;
		} else if (!read_bit(PINB, 5) && read_flag(TRAINING_MODE)) {
			clear_flag(TRAINING_MODE);
			LCDclr();
			sprintf(lcd_buffer,"Testing Mode!");	                 
			LCDGotoXY(1,1);
			LCDstring(lcd_buffer, strlen(lcd_buffer));
			message_print_timer = MESSAGE_PRINT_MS;
		}

		// ----- sound detected LED ------
		if (read_flag(RECENT_SOUND)) LED_on;
		else LED_off; 
		// -------------------------------

		// -------------------- SAMPLE PROCESSING ---------------------------------------
		if (read_flag(BUTTON_PUSHED) && read_flag(RECENT_SOUND) && !read_flag(SAMPLES_PROCESSED)) {
				// ----- STORE COEFF / TAKE MORE SAMPLES -----
				while (stored_coeffs < NUM_TRAINING_FRAMES) {
					process_samples();
					for (int i=0; i<NUM_COEFF; i++) {
						for (int j=0; j<NUM_FRAMES; j++) {
							coeff[i][j+stored_coeffs] = frame[i][j];
						}
					}
					stored_coeffs += NUM_FRAMES;
				}
				unsigned int num_zeros = 0;
				for(unsigned int j = 0; j < NUM_COEFF; j++) {
					for(unsigned int k = 0; k < NUM_TRAINING_FRAMES; k++) {
						if (coeff[j][k] == 0) {
							num_zeros++;
						}
					}
				}
				if (num_zeros > 4000) {
					fprintf(stdout,"TOO MANY ZEROS\r\n");
					LCDclr();
					sprintf(lcd_buffer,"Longer speech");	                 
					LCDGotoXY(0,0);
					LCDstring(lcd_buffer, strlen(lcd_buffer));
					sprintf(lcd_buffer,"sample required");	                 
					LCDGotoXY(0,1);
					LCDstring(lcd_buffer, strlen(lcd_buffer));
					sprintf(lcd_buffer,"OK");	                 
					LCDGotoXY(10,3);
					LCDstring(lcd_buffer, strlen(lcd_buffer));
					while (read_bit(PINB, 0));
					set_flag(PRINT_WELCOME);
					stored_coeffs = 0;
					continue;
				}
				set_flag(SAMPLES_PROCESSED);
				set_bit(PORTA, 2);
				// --------------------------------------------
				getCentroids();
				if (!read_flag(TRAINING_MODE)) {
					recognize();
					LCDclr();
					unsigned int num_outputs = 3;
					sprintf(lcd_buffer,"Proximities:    ");
					LCDGotoXY(0,0);
					LCDstring(lcd_buffer, strlen(lcd_buffer));
					if (samples_entered>0 && samples_entered<3) num_outputs = samples_entered;
					for (int i=0; i<num_outputs; i++) {
						if (i==0) LCDGotoXY(0,1);
						else if (i==1) LCDGotoXY(-4,2);
						else LCDGotoXY(-4,3);
						if (speaker_dist[i]<500) {
							sprintf(lcd_buffer,"Match: %i", speaker_dist[i]);
						} else {
							sprintf(lcd_buffer,"Miss:  %i", speaker_dist[i]);
						}
						LCDstring(lcd_buffer, strlen(lcd_buffer));
					}              
					LCDGotoXY(10,3);
					sprintf(lcd_buffer,"OK");
					LCDstring(lcd_buffer, strlen(lcd_buffer));
					while (read_bit(PINB, 0));
				} else {
					sample_num = mod(samples_entered,3);
					for(unsigned int j = 0; j < NUM_COEFF; j++) {
						for(unsigned int k = 0; k < NUM_CENTROIDS; k++) {
							speakers[sample_num][j][k] = c[j][k];
						}
					}
					samples_entered++;
					fprintf(stdout,"Stored as:\r\n");
					for(unsigned int j = 0; j < NUM_COEFF; j++) {
						fprintf(stdout,"{");
						for(unsigned int k = 0; k < NUM_CENTROIDS; k++) {
							fprintf(stdout,"%i, ",speakers[sample_num][j][k]);
						}
						fprintf(stdout,"},\r\n");
					}
					if (samples_entered == 1) {
						clear_bit(PORTA, 4);
						clear_bit(PORTA, 5);
					} else if (samples_entered == 2) {
						set_bit(PORTA, 5);
					} else if (samples_entered == 3) {
						set_bit(PORTA, 4);
					}
				}
				// ---------------------------------------------
			set_flag(PRINT_WELCOME);
			clear_flag(SAMPLES_PROCESSED);
			stored_coeffs = 0;
		}
		// ------------------END OF SAMPLE PROCESSING BLOCK ----------------
	}
end 
